/************************************************************************
*  TestOpt.C (Simple Test Program for Minimization Algorithm)
*
*  Version:                     1.6
*  Date:                        11/30/94
*  Author:                      O. Knueppel
*************************************************************************/

#include "uconsopt.h"
#include "func.h"

//
// This simple test program displays how to use different subroutines
// for the computation of the function/derivative values. For a
// more simple, but also slower approach see "SimpleTestOpt.C".
//

//
// The Rastrigin function [Rastrigin 1974] is used as test function.
// This function has one global minimum:
//
//  x = (0,0), y = -2
//

VOID SetTestDomain (INTERVAL_VECTOR & x)
// The domain of the test function
{
  Resize (x, 2);
  x(1) = Hull (-1.0, 1.0);
  x(2) = Hull (-1.0, 1.0);
}

REAL RealTestFunction (VECTOR & x)
// real valued test function
{
  return Sqr (x(1)) + Sqr (x(2)) - Cos (18.0 * x(1)) - Cos (18.0 * x(2));
}

INTERVAL IntervalTestFunction (INTERVAL_VECTOR & x)
// interval valued test function
{
  return Sqr (x(1)) + Sqr (x(2)) - Cos (18.0 * x(1)) - Cos (18.0 * x(2));
}

INTERVAL_VECTOR IntervalTestGradient (INTERVAL_VECTOR & x)
// interval valued gradient of test function
{
  INTERVAL_VECTOR t(2);

  t(1) = 2.0 * x(1) + 18.0 * Sin (18.0 * x(1));
  t(2) = 2.0 * x(2) + 18.0 * Sin (18.0 * x(2));
  return t;
}

INTERVAL_MATRIX IntervalTestHessian (INTERVAL_VECTOR & x)
// interval valued Hessian of test function
{
  INTERVAL_MATRIX t(2,2);

  t(1,1) = 2.0 + 324.0 * Cos (18.0 * x(1)); t(1,2) = 0.0;
  t(2,1) = 0.0; t(2,2) = 2.0 + 324.0 * Cos (18.0 * x(2));
  return t;
}

INT main()
{
  INT Iterations, BranchLevels;
  SOLUTIONLIST SolutionList;
  APPROXIMATIONLIST ApproximationList;
  INTERVAL_VECTOR TestDomain;
  REAL LowerBound, UpperBound;

  SetTestDomain (TestDomain);

  cout << "Computing all global minima of the Rastrigin function" << endl << endl;

  cout << "nit = "; cin >> Iterations;
  cout << "nd  = "; cin >> BranchLevels;

  StartUnconstrainedOptimization (SolutionList, ApproximationList,
				  Iterations, BranchLevels, 1E-6, 0.2, 0.2,
				  LowerBound, UpperBound, TestDomain,
				  RealTestFunction, IntervalTestFunction,
				  IntervalTestGradient, IntervalTestHessian);

  CleanUpLists (SolutionList, ApproximationList, 1e-6, LowerBound, UpperBound);

  cout << "f min in " << Hull (LowerBound, UpperBound) << endl;
  cout << "Solution List:" << endl << SolutionList << endl;
  cout << "Approximation List:" << endl << ApproximationList << endl;
  return 0;
}
