/************************************************************************
*  SimpleOptDriver.C (Simple Driver for Minimization Algorithm)
*
*  Version:                     1.2
*  Date:                        11/30/94
*  Author:                      O. Knueppel
*************************************************************************/

#include "soptdrv.h"
#include "uconsopt.h"
#include "func.h"
#include "iautodif.h"

//
// This simple driver for the minimization algorithm uses
// automatic differentiation to compute the derivatives.
// Because the size of this driver program should be kept small,
// all real valued function/derivative evaluations are obtained
// from an interval evaluation of the same function/derivative.
// Additionally, for function evaluations, the gradient is always
// computed but not used. Therefore, the is much speed penalty for
// this approach. For a more sophisticated approach see "TestOpt.C"
// together with "UnconstrainedOpt.C".
//

INTERVAL_AUTODIFF (*Function)(INTERVAL_VECTOR &);

static INTERVAL TheIntervalFunction (INTERVAL_VECTOR & x)
{
  INTERVAL_AUTODIFF::ComputeHessian = 0;
  return FunctionValue (Function (x));
}

static REAL TheRealFunction (VECTOR & x)
{
  return Mid (TheIntervalFunction (Hull (x)));
}

static INTERVAL_VECTOR TheIntervalGradient (INTERVAL_VECTOR & x)
{
  INTERVAL_AUTODIFF::ComputeHessian = 0;
  return GradientValue (Function (x));
}

static VECTOR TheRealGradient (VECTOR & x)
{
  return Mid (TheIntervalGradient (Hull (x)));
}

static INTERVAL_MATRIX TheIntervalHessian (INTERVAL_VECTOR & x)
{
  INTERVAL_AUTODIFF::ComputeHessian = 1;
  return HessianValue (Function (x));
}

static MATRIX TheRealHessian (VECTOR & x)
{
  return Mid (TheIntervalHessian (Hull (x)));
}

VOID SimpleOptDriver (INT Iterations, INT BranchLevels,
		      INTERVAL_AUTODIFF (*TheFunction)(INTERVAL_VECTOR &),
		      INTERVAL_VECTOR TestDomain,
		      SOLUTIONLIST & SolutionList, APPROXIMATIONLIST & ApproximationList,
		      INTERVAL & ValueBounds,
		      REAL Tolerance, REAL RaiseFactor, REAL HeuristicExpandFactor)
{
  REAL LowerBound, UpperBound;

  Function = TheFunction;
  StartUnconstrainedOptimization (SolutionList, ApproximationList,
				  Iterations, BranchLevels, Tolerance,
				  RaiseFactor, HeuristicExpandFactor,
				  LowerBound, UpperBound, TestDomain,
				  TheRealFunction, TheIntervalFunction,
				  TheIntervalGradient, TheIntervalHessian);

  CleanUpLists (SolutionList, ApproximationList, Tolerance, LowerBound, UpperBound);
  ValueBounds = Hull (LowerBound, UpperBound);
}

